@@ -11,6 +11,7 @@
 #include "iconloader_p.h"
 #include "penwidthutil_p.h"
 #include "scenegeometry_p.h"
+#include "openinghourscache_p.h"
 #include "../style/mapcssdeclaration_p.h"
 #include "../style/mapcssstate_p.h"
 #include "../style/mapcssresult_p.h"
@@ -33,10 +34,10 @@ namespace KOSMIndoorMap {
 class SceneControllerPrivate
 {
 public:
-    const MapData *m_data = nullptr;
+    MapData m_data;
     const MapCSSStyle *m_styleSheet = nullptr;
     const View *m_view = nullptr;
-    std::vector<OverlaySource> m_overlaySources;
+    std::vector<QPointer<AbstractOverlaySource>> m_overlaySources;
     mutable std::vector<OSM::Element> m_hiddenElements;
 
     MapCSSResult m_styleResult;
@@ -44,6 +45,7 @@ public:
     QFont m_defaultFont;
     QPolygonF m_labelPlacementPath;
     IconLoader m_iconLoader;
+    OpeningHoursCache m_openingHours;
 
     OSM::TagKey m_layerTag;
     OSM::TagKey m_typeTag;
@@ -58,15 +60,17 @@ using namespace KOSMIndoorMap;
 SceneController::SceneController() : d(new SceneControllerPrivate) {}
 SceneController::~SceneController() = default;
 
-void SceneController::setDataSet(const MapData *data)
+void SceneController::setMapData(const MapData &data)
 {
     d->m_data = data;
-    if (d->m_data) {
-        d->m_layerTag = data->dataSet().tagKey("layer");
-        d->m_typeTag = data->dataSet().tagKey("type");
+    if (!d->m_data.isEmpty()) {
+        d->m_layerTag = data.dataSet().tagKey("layer");
+        d->m_typeTag = data.dataSet().tagKey("type");
+        d->m_openingHours.setMapData(data);
     } else {
         d->m_layerTag = {};
         d->m_typeTag = {};
+        d->m_openingHours.setMapData(MapData());
     }
     d->m_dirty = true;
 }
@@ -80,10 +84,11 @@ void SceneController::setStyleSheet(const MapCSSStyle *styleSheet)
 void SceneController::setView(const View *view)
 {
     d->m_view = view;
+    QObject::connect(view, &View::timeChanged, [this]() { d->m_dirty = true; });
     d->m_dirty = true;
 }
 
-void SceneController::setOverlaySources(std::vector<OverlaySource> &&overlays)
+void SceneController::setOverlaySources(std::vector<QPointer<AbstractOverlaySource>> &&overlays)
 {
     d->m_overlaySources = std::move(overlays);
     d->m_dirty = true;
@@ -111,32 +116,33 @@ void SceneController::updateScene(SceneGraph &sg) const
     }
     sg.setZoomLevel(d->m_view->zoomLevel());
     sg.setCurrentFloorLevel(d->m_view->level());
+    d->m_openingHours.setTimeRange(d->m_view->beginTime(), d->m_view->endTime());
     d->m_dirty = false;
 
     sg.beginSwap();
     updateCanvas(sg);
 
-    if (!d->m_data) { // if we don't have map data yet, we just need to get canvas styling here
+    if (d->m_data.isEmpty()) { // if we don't have map data yet, we just need to get canvas styling here
         sg.endSwap();
         return;
     }
 
     // find all intermediate levels below or above the currently selected "full" level
-    auto it = d->m_data->levelMap().find(MapLevel(d->m_view->level()));
-    if (it == d->m_data->levelMap().end()) {
+    auto it = d->m_data.levelMap().find(MapLevel(d->m_view->level()));
+    if (it == d->m_data.levelMap().end()) {
         return;
     }
 
     auto beginIt = it;
-    if (beginIt != d->m_data->levelMap().begin()) {
+    if (beginIt != d->m_data.levelMap().begin()) {
         do {
             --beginIt;
-        } while (!(*beginIt).first.isFullLevel() && beginIt != d->m_data->levelMap().begin());
+        } while (!(*beginIt).first.isFullLevel() && beginIt != d->m_data.levelMap().begin());
         ++beginIt;
     }
 
     auto endIt = it;
-    for (++endIt; endIt != d->m_data->levelMap().end(); ++endIt) {
+    for (++endIt; endIt != d->m_data.levelMap().end(); ++endIt) {
         if ((*endIt).first.isFullLevel()) {
             break;
         }
@@ -145,7 +151,7 @@ void SceneController::updateScene(SceneGraph &sg) const
     // collect elements that the overlay want to hide
     d->m_hiddenElements.clear();
     for (const auto &overlaySource : d->m_overlaySources) {
-        overlaySource.hiddenElements(d->m_hiddenElements);
+        overlaySource->hiddenElements(d->m_hiddenElements);
     }
     std::sort(d->m_hiddenElements.begin(), d->m_hiddenElements.end());
 
@@ -162,7 +168,7 @@ void SceneController::updateScene(SceneGraph &sg) const
     // update overlay elements
     d->m_overlay = true;
     for (const auto &overlaySource : d->m_overlaySources) {
-        overlaySource.forEach(d->m_view->level(), [this, &geoBbox, &sg](OSM::Element e, int floorLevel) {
+        overlaySource->forEach(d->m_view->level(), [this, &geoBbox, &sg](OSM::Element e, int floorLevel) {
             if (OSM::intersects(geoBbox, e.boundingBox()) && e.type() != OSM::Type::Null) {
                 updateElement(e, floorLevel, sg);
             }
@@ -206,6 +212,7 @@ void SceneController::updateElement(OSM::Element e, int level, SceneGraph &sg) c
     state.element = e;
     state.zoomLevel = d->m_view->zoomLevel();
     state.floorLevel = d->m_view->level();
+    state.openingHours = &d->m_openingHours;
     d->m_styleSheet->evaluate(state, d->m_styleResult);
 
     if (d->m_styleResult.hasAreaProperties()) {
@@ -450,11 +457,29 @@ void SceneController::updateElement(OSM::Element e, int level, SceneGraph &sg) c
 
 QPolygonF SceneController::createPolygon(OSM::Element e) const
 {
-    const auto path = e.outerPath(d->m_data->dataSet());
+    const auto path = e.outerPath(d->m_data.dataSet());
+    if (path.empty()) {
+        return {};
+    }
+
     QPolygonF poly;
-    poly.reserve(path.size());
-    for (auto node : path) {
-        poly.push_back(d->m_view->mapGeoToScene(node->coordinate));
+    // Element::outerPath takes care of re-assembling broken up line segments
+    // the below takes care of properly merging broken up polygons
+    for (auto it = path.begin(); it != path.end();) {
+        QPolygonF subPoly;
+        subPoly.reserve(path.size());
+        OSM::Id pathBegin = (*it)->id;
+
+        auto subIt = it;
+        for (; subIt != path.end(); ++subIt) {
+            subPoly.push_back(d->m_view->mapGeoToScene((*subIt)->coordinate));
+            if ((*subIt)->id == pathBegin && subIt != it && subIt != std::prev(path.end())) {
+                ++subIt;
+                break;
+            }
+        }
+        it = subIt;
+        poly = poly.isEmpty() ? std::move(subPoly) : poly.united(subPoly);
     }
     return poly;
 }
@@ -472,7 +497,7 @@ QPainterPath SceneController::createPath(const OSM::Element e, QPolygonF &outerP
         if (mem.type() != OSM::Type::Way || !isInner) {
             continue;
         }
-        if (auto way = d->m_data->dataSet().way(mem.id)) {
+        if (auto way = d->m_data.dataSet().way(mem.id)) {
             const auto subPoly = createPolygon(OSM::Element(way));
             QPainterPath subPath;
             subPath.addPolygon(subPoly);
